# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql.base import PGCompiler
from sqlalchemy.sql import selectable
from sqlalchemy.sql.crud import (REQUIRED, _create_bind_param,
                                 _extend_values_for_multiparams,
                                 _get_stmt_parameter_tuples_params,
                                 _get_update_multitable_params,
                                 _key_getters_for_crud_column, _scan_cols,
                                 _scan_insert_from_select_cols)

from crate.client.sqlalchemy.compiler import CrateCompiler


class CrateCompilerSA14(CrateCompiler):

    def returning_clause(self, stmt, returning_cols):
        """
        Generate RETURNING clause, PostgreSQL-compatible.
        """
        return PGCompiler.returning_clause(self, stmt, returning_cols)

    def visit_update(self, update_stmt, **kw):

        compile_state = update_stmt._compile_state_factory(
            update_stmt, self, **kw
        )
        update_stmt = compile_state.statement

        # [14] CrateDB patch.
        if not compile_state._dict_parameters and \
                not hasattr(update_stmt, '_crate_specific'):
            return super().visit_update(update_stmt, **kw)

        toplevel = not self.stack
        if toplevel:
            self.isupdate = True
            if not self.compile_state:
                self.compile_state = compile_state

        extra_froms = compile_state._extra_froms
        is_multitable = bool(extra_froms)

        if is_multitable:
            # main table might be a JOIN
            main_froms = set(selectable._from_objects(update_stmt.table))
            render_extra_froms = [
                f for f in extra_froms if f not in main_froms
            ]
            correlate_froms = main_froms.union(extra_froms)
        else:
            render_extra_froms = []
            correlate_froms = {update_stmt.table}

        self.stack.append(
            {
                "correlate_froms": correlate_froms,
                "asfrom_froms": correlate_froms,
                "selectable": update_stmt,
            }
        )

        text = "UPDATE "

        if update_stmt._prefixes:
            text += self._generate_prefixes(
                update_stmt, update_stmt._prefixes, **kw
            )

        table_text = self.update_tables_clause(
            update_stmt, update_stmt.table, render_extra_froms, **kw
        )

        # [14] CrateDB patch.
        crud_params = _get_crud_params(
            self, update_stmt, compile_state, **kw
        )

        if update_stmt._hints:
            dialect_hints, table_text = self._setup_crud_hints(
                update_stmt, table_text
            )
        else:
            dialect_hints = None

        if update_stmt._independent_ctes:
            for cte in update_stmt._independent_ctes:
                cte._compiler_dispatch(self, **kw)

        text += table_text

        text += " SET "

        # [14] CrateDB patch begin.
        include_table = \
            extra_froms and self.render_table_with_column_in_update_from

        set_clauses = []

        for c, expr, value in crud_params:
            key = c._compiler_dispatch(self, include_table=include_table)
            clause = key + ' = ' + value
            set_clauses.append(clause)

        for k, v in compile_state._dict_parameters.items():
            if isinstance(k, str) and '[' in k:
                bindparam = sa.sql.bindparam(k, v)
                clause = k + ' = ' + self.process(bindparam)
                set_clauses.append(clause)

        text += ', '.join(set_clauses)
        # [14] CrateDB patch end.

        if self.returning or update_stmt._returning:
            if self.returning_precedes_values:
                text += " " + self.returning_clause(
                    update_stmt, self.returning or update_stmt._returning
                )

        if extra_froms:
            extra_from_text = self.update_from_clause(
                update_stmt,
                update_stmt.table,
                render_extra_froms,
                dialect_hints,
                **kw
            )
            if extra_from_text:
                text += " " + extra_from_text

        if update_stmt._where_criteria:
            t = self._generate_delimited_and_list(
                update_stmt._where_criteria, **kw
            )
            if t:
                text += " WHERE " + t

        limit_clause = self.update_limit_clause(update_stmt)
        if limit_clause:
            text += " " + limit_clause

        if (
                self.returning or update_stmt._returning
        ) and not self.returning_precedes_values:
            text += " " + self.returning_clause(
                update_stmt, self.returning or update_stmt._returning
            )

        if self.ctes:
            nesting_level = len(self.stack) if not toplevel else None
            text = self._render_cte_clause(nesting_level=nesting_level) + text

        self.stack.pop(-1)

        return text


def _get_crud_params(compiler, stmt, compile_state, **kw):
    """create a set of tuples representing column/string pairs for use
    in an INSERT or UPDATE statement.

    Also generates the Compiled object's postfetch, prefetch, and
    returning column collections, used for default handling and ultimately
    populating the CursorResult's prefetch_cols() and postfetch_cols()
    collections.

    """

    compiler.postfetch = []
    compiler.insert_prefetch = []
    compiler.update_prefetch = []
    compiler.returning = []

    # getters - these are normally just column.key,
    # but in the case of mysql multi-table update, the rules for
    # .key must conditionally take tablename into account
    (
        _column_as_key,
        _getattr_col_key,
        _col_bind_name,
    ) = getters = _key_getters_for_crud_column(compiler, stmt, compile_state)

    compiler._key_getters_for_crud_column = getters

    # no parameters in the statement, no parameters in the
    # compiled params - return binds for all columns
    if compiler.column_keys is None and compile_state._no_parameters:
        return [
            (
                c,
                compiler.preparer.format_column(c),
                _create_bind_param(compiler, c, None, required=True),
            )
            for c in stmt.table.columns
        ]

    if compile_state._has_multi_parameters:
        spd = compile_state._multi_parameters[0]
        stmt_parameter_tuples = list(spd.items())
    elif compile_state._ordered_values:
        spd = compile_state._dict_parameters
        stmt_parameter_tuples = compile_state._ordered_values
    elif compile_state._dict_parameters:
        spd = compile_state._dict_parameters
        stmt_parameter_tuples = list(spd.items())
    else:
        stmt_parameter_tuples = spd = None

    # if we have statement parameters - set defaults in the
    # compiled params
    if compiler.column_keys is None:
        parameters = {}
    elif stmt_parameter_tuples:
        parameters = dict(
            (_column_as_key(key), REQUIRED)
            for key in compiler.column_keys
            if key not in spd
        )
    else:
        parameters = dict(
            (_column_as_key(key), REQUIRED) for key in compiler.column_keys
        )

    # create a list of column assignment clauses as tuples
    values = []

    if stmt_parameter_tuples is not None:
        _get_stmt_parameter_tuples_params(
            compiler,
            compile_state,
            parameters,
            stmt_parameter_tuples,
            _column_as_key,
            values,
            kw,
        )

    check_columns = {}

    # special logic that only occurs for multi-table UPDATE
    # statements
    if compile_state.isupdate and compile_state.is_multitable:
        _get_update_multitable_params(
            compiler,
            stmt,
            compile_state,
            stmt_parameter_tuples,
            check_columns,
            _col_bind_name,
            _getattr_col_key,
            values,
            kw,
        )

    if compile_state.isinsert and stmt._select_names:
        _scan_insert_from_select_cols(
            compiler,
            stmt,
            compile_state,
            parameters,
            _getattr_col_key,
            _column_as_key,
            _col_bind_name,
            check_columns,
            values,
            kw,
        )
    else:
        _scan_cols(
            compiler,
            stmt,
            compile_state,
            parameters,
            _getattr_col_key,
            _column_as_key,
            _col_bind_name,
            check_columns,
            values,
            kw,
        )

    # [14] CrateDB patch.
    #
    # This sanity check performed by SQLAlchemy currently needs to be
    # deactivated in order to satisfy the rewriting logic of the CrateDB
    # dialect in `rewrite_update` and `visit_update`.
    #
    # It can be quickly reproduced by activating this section and running the
    # test cases::
    #
    #   ./bin/test -vvvv -t dict_test
    #
    # That croaks like::
    #
    #   sqlalchemy.exc.CompileError: Unconsumed column names: characters_name, data['nested']
    #
    # TODO: Investigate why this is actually happening and eventually mitigate
    #       the root cause.
    """
    if parameters and stmt_parameter_tuples:
        check = (
            set(parameters)
            .intersection(_column_as_key(k) for k, v in stmt_parameter_tuples)
            .difference(check_columns)
        )
        if check:
            raise exc.CompileError(
                "Unconsumed column names: %s"
                % (", ".join("%s" % (c,) for c in check))
            )
    """

    if compile_state._has_multi_parameters:
        values = _extend_values_for_multiparams(
            compiler,
            stmt,
            compile_state,
            values,
            _column_as_key,
            kw,
        )
    elif (
            not values
            and compiler.for_executemany  # noqa: W503
            and compiler.dialect.supports_default_metavalue  # noqa: W503
    ):
        # convert an "INSERT DEFAULT VALUES"
        # into INSERT (firstcol) VALUES (DEFAULT) which can be turned
        # into an in-place multi values.  This supports
        # insert_executemany_returning mode :)
        values = [
            (
                stmt.table.columns[0],
                compiler.preparer.format_column(stmt.table.columns[0]),
                "DEFAULT",
            )
        ]

    return values



class ExtraAssertions:
    """
    Additional assert methods for unittest.

    - https://github.com/python/cpython/issues/71339
    - https://bugs.python.org/issue14819
    - https://bugs.python.org/file43047/extra_assertions.patch
    """

    def assertIsSubclass(self, cls, superclass, msg=None):
        try:
            r = issubclass(cls, superclass)
        except TypeError:
            if not isinstance(cls, type):
                self.fail(self._formatMessage(msg,
                          '%r is not a class' % (cls,)))
            raise
        if not r:
            self.fail(self._formatMessage(msg,
                      '%r is not a subclass of %r' % (cls, superclass)))



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import datetime
from ipaddress import IPv4Address
from unittest import TestCase
from unittest.mock import MagicMock
try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo

import pytz

from crate.client import connect
from crate.client.converter import DataType, DefaultTypeConverter
from crate.client.http import Client
from crate.client.test_util import ClientMocked


class CursorTest(TestCase):

    @staticmethod
    def get_mocked_connection():
        client = MagicMock(spec=Client)
        return connect(client=client)

    def test_create_with_timezone_as_datetime_object(self):
        """
        Verify the cursor returns timezone-aware `datetime` objects when requested to.
        Switching the time zone at runtime on the cursor object is possible.
        Here: Use a `datetime.timezone` instance.
        """

        connection = self.get_mocked_connection()

        tz_mst = datetime.timezone(datetime.timedelta(hours=7), name="MST")
        cursor = connection.cursor(time_zone=tz_mst)

        self.assertEqual(cursor.time_zone.tzname(None), "MST")
        self.assertEqual(cursor.time_zone.utcoffset(None), datetime.timedelta(seconds=25200))

        cursor.time_zone = datetime.timezone.utc
        self.assertEqual(cursor.time_zone.tzname(None), "UTC")
        self.assertEqual(cursor.time_zone.utcoffset(None), datetime.timedelta(0))

    def test_create_with_timezone_as_pytz_object(self):
        """
        Verify the cursor returns timezone-aware `datetime` objects when requested to.
        Here: Use a `pytz.timezone` instance.
        """
        connection = self.get_mocked_connection()
        cursor = connection.cursor(time_zone=pytz.timezone('Australia/Sydney'))
        self.assertEqual(cursor.time_zone.tzname(None), "Australia/Sydney")

        # Apparently, when using `pytz`, the timezone object does not return an offset.
        # Nevertheless, it works, as demonstrated per doctest in `cursor.txt`.
        self.assertEqual(cursor.time_zone.utcoffset(None), None)

    def test_create_with_timezone_as_zoneinfo_object(self):
        """
        Verify the cursor returns timezone-aware `datetime` objects when requested to.
        Here: Use a `zoneinfo.ZoneInfo` instance.
        """
        connection = self.get_mocked_connection()
        cursor = connection.cursor(time_zone=zoneinfo.ZoneInfo('Australia/Sydney'))
        self.assertEqual(cursor.time_zone.key, 'Australia/Sydney')

    def test_create_with_timezone_as_utc_offset_success(self):
        """
        Verify the cursor returns timezone-aware `datetime` objects when requested to.
        Here: Use a UTC offset in string format.
        """
        connection = self.get_mocked_connection()
        cursor = connection.cursor(time_zone="+0530")
        self.assertEqual(cursor.time_zone.tzname(None), "+0530")
        self.assertEqual(cursor.time_zone.utcoffset(None), datetime.timedelta(seconds=19800))

        connection = self.get_mocked_connection()
        cursor = connection.cursor(time_zone="-1145")
        self.assertEqual(cursor.time_zone.tzname(None), "-1145")
        self.assertEqual(cursor.time_zone.utcoffset(None), datetime.timedelta(days=-1, seconds=44100))

    def test_create_with_timezone_as_utc_offset_failure(self):
        """
        Verify the cursor croaks when trying to create it with invalid UTC offset strings.
        """
        connection = self.get_mocked_connection()
        with self.assertRaises(AssertionError) as ex:
            connection.cursor(time_zone="foobar")
        self.assertEqual(str(ex.exception), "Time zone 'foobar' is given in invalid UTC offset format")

        connection = self.get_mocked_connection()
        with self.assertRaises(ValueError) as ex:
            connection.cursor(time_zone="+abcd")
        self.assertEqual(str(ex.exception), "Time zone '+abcd' is given in invalid UTC offset format: "
                                            "invalid literal for int() with base 10: '+ab'")

    def test_create_with_timezone_connection_cursor_precedence(self):
        """
        Verify that the time zone specified on the cursor object instance
        takes precedence over the one specified on the connection instance.
        """
        client = MagicMock(spec=Client)
        connection = connect(client=client, time_zone=pytz.timezone('Australia/Sydney'))
        cursor = connection.cursor(time_zone="+0530")
        self.assertEqual(cursor.time_zone.tzname(None), "+0530")
        self.assertEqual(cursor.time_zone.utcoffset(None), datetime.timedelta(seconds=19800))

    def test_execute_with_args(self):
        client = MagicMock(spec=Client)
        conn = connect(client=client)
        c = conn.cursor()
        statement = 'select * from locations where position = ?'
        c.execute(statement, 1)
        client.sql.assert_called_once_with(statement, 1, None)
        conn.close()

    def test_execute_with_bulk_args(self):
        client = MagicMock(spec=Client)
        conn = connect(client=client)
        c = conn.cursor()
        statement = 'select * from locations where position = ?'
        c.execute(statement, bulk_parameters=[[1]])
        client.sql.assert_called_once_with(statement, None, [[1]])
        conn.close()

    def test_execute_with_converter(self):
        client = ClientMocked()
        conn = connect(client=client)

        # Use the set of data type converters from `DefaultTypeConverter`
        # and add another custom converter.
        converter = DefaultTypeConverter(
            {DataType.BIT: lambda value: value is not None and int(value[2:-1], 2) or None})

        # Create a `Cursor` object with converter.
        c = conn.cursor(converter=converter)

        # Make up a response using CrateDB data types `TEXT`, `IP`,
        # `TIMESTAMP`, `BIT`.
        conn.client.set_next_response({
            "col_types": [4, 5, 11, 25],
            "cols": ["name", "address", "timestamp", "bitmask"],
            "rows": [
                ["foo", "10.10.10.1", 1658167836758, "B'0110'"],
                [None, None, None, None],
            ],
            "rowcount": 1,
            "duration": 123
        })

        c.execute("")
        result = c.fetchall()
        self.assertEqual(result, [
            ['foo', IPv4Address('10.10.10.1'), datetime.datetime(2022, 7, 18, 18, 10, 36, 758000), 6],
            [None, None, None, None],
        ])

        conn.close()

    def test_execute_with_converter_and_invalid_data_type(self):
        client = ClientMocked()
        conn = connect(client=client)
        converter = DefaultTypeConverter()

        # Create a `Cursor` object with converter.
        c = conn.cursor(converter=converter)

        # Make up a response using CrateDB data types `TEXT`, `IP`,
        # `TIMESTAMP`, `BIT`.
        conn.client.set_next_response({
            "col_types": [999],
            "cols": ["foo"],
            "rows": [
                ["n/a"],
            ],
            "rowcount": 1,
            "duration": 123
        })

        c.execute("")
        with self.assertRaises(ValueError) as ex:
            c.fetchone()
        self.assertEqual(ex.exception.args, ("999 is not a valid DataType",))

    def test_execute_array_with_converter(self):
        client = ClientMocked()
        conn = connect(client=client)
        converter = DefaultTypeConverter()
        cursor = conn.cursor(converter=converter)

        conn.client.set_next_response({
            "col_types": [4, [100, 5]],
            "cols": ["name", "address"],
            "rows": [["foo", ["10.10.10.1", "10.10.10.2"]]],
            "rowcount": 1,
            "duration": 123
        })

        cursor.execute("")
        result = cursor.fetchone()
        self.assertEqual(result, [
            'foo',
            [IPv4Address('10.10.10.1'), IPv4Address('10.10.10.2')],
        ])

    def test_execute_array_with_converter_and_invalid_collection_type(self):
        client = ClientMocked()
        conn = connect(client=client)
        converter = DefaultTypeConverter()
        cursor = conn.cursor(converter=converter)

        # Converting collections only works for `ARRAY`s. (ID=100).
        # When using `DOUBLE` (ID=6), it should croak.
        conn.client.set_next_response({
            "col_types": [4, [6, 5]],
            "cols": ["name", "address"],
            "rows": [["foo", ["10.10.10.1", "10.10.10.2"]]],
            "rowcount": 1,
            "duration": 123
        })

        cursor.execute("")

        with self.assertRaises(ValueError) as ex:
            cursor.fetchone()
        self.assertEqual(ex.exception.args, ("Data type 6 is not implemented as collection type",))

    def test_execute_nested_array_with_converter(self):
        client = ClientMocked()
        conn = connect(client=client)
        converter = DefaultTypeConverter()
        cursor = conn.cursor(converter=converter)

        conn.client.set_next_response({
            "col_types": [4, [100, [100, 5]]],
            "cols": ["name", "address_buckets"],
            "rows": [["foo", [["10.10.10.1", "10.10.10.2"], ["10.10.10.3"], [], None]]],
            "rowcount": 1,
            "duration": 123
        })

        cursor.execute("")
        result = cursor.fetchone()
        self.assertEqual(result, [
            'foo',
            [[IPv4Address('10.10.10.1'), IPv4Address('10.10.10.2')], [IPv4Address('10.10.10.3')], [], None],
        ])

    def test_executemany_with_converter(self):
        client = ClientMocked()
        conn = connect(client=client)
        converter = DefaultTypeConverter()
        cursor = conn.cursor(converter=converter)

        conn.client.set_next_response({
            "col_types": [4, 5],
            "cols": ["name", "address"],
            "rows": [["foo", "10.10.10.1"]],
            "rowcount": 1,
            "duration": 123
        })

        cursor.executemany("", [])
        result = cursor.fetchall()

        # ``executemany()`` is not intended to be used with statements returning result
        # sets. The result will always be empty.
        self.assertEqual(result, [])

    def test_execute_with_timezone(self):
        client = ClientMocked()
        conn = connect(client=client)

        # Create a `Cursor` object with `time_zone`.
        tz_mst = datetime.timezone(datetime.timedelta(hours=7), name="MST")
        c = conn.cursor(time_zone=tz_mst)

        # Make up a response using CrateDB data type `TIMESTAMP`.
        conn.client.set_next_response({
            "col_types": [4, 11],
            "cols": ["name", "timestamp"],
            "rows": [
                ["foo", 1658167836758],
                [None, None],
            ],
        })

        # Run execution and verify the returned `datetime` object is timezone-aware,
        # using the designated timezone object.
        c.execute("")
        result = c.fetchall()
        self.assertEqual(result, [
            [
                'foo',
                datetime.datetime(2022, 7, 19, 1, 10, 36, 758000,
                                  tzinfo=datetime.timezone(datetime.timedelta(seconds=25200), 'MST')),
            ],
            [
                None,
                None,
            ],
        ])
        self.assertEqual(result[0][1].tzname(), "MST")

        # Change timezone and verify the returned `datetime` object is using it.
        c.time_zone = datetime.timezone.utc
        c.execute("")
        result = c.fetchall()
        self.assertEqual(result, [
            [
                'foo',
                datetime.datetime(2022, 7, 18, 18, 10, 36, 758000, tzinfo=datetime.timezone.utc),
            ],
            [
                None,
                None,
            ],
        ])
        self.assertEqual(result[0][1].tzname(), "UTC")

        conn.close()



import datetime

from urllib3 import Timeout

from .connection import Connection
from .http import Client
from crate.client import connect
from unittest import TestCase

from ..testing.settings import crate_host


class ConnectionTest(TestCase):

    def test_connection_mock(self):
        """
        For testing purposes it is often useful to replace the client used for
        communication with the CrateDB server with a stub or mock.

        This can be done by passing an object of the Client class when calling the
        ``connect`` method.
        """

        class MyConnectionClient:
            active_servers = ["localhost:4200"]

            def __init__(self):
                pass

            def server_infos(self, server):
                return ("localhost:4200", "my server", "0.42.0")

        connection = connect([crate_host], client=MyConnectionClient())
        self.assertIsInstance(connection, Connection)
        self.assertEqual(connection.client.server_infos("foo"), ('localhost:4200', 'my server', '0.42.0'))

    def test_lowest_server_version(self):
        infos = [(None, None, '0.42.3'),
                 (None, None, '0.41.8'),
                 (None, None, 'not a version')]

        client = Client(servers="localhost:4200 localhost:4201 localhost:4202")
        client.server_infos = lambda server: infos.pop()
        connection = connect(client=client)
        self.assertEqual((0, 41, 8), connection.lowest_server_version.version)
        connection.close()

    def test_invalid_server_version(self):
        client = Client(servers="localhost:4200")
        client.server_infos = lambda server: (None, None, "No version")
        connection = connect(client=client)
        self.assertEqual((0, 0, 0), connection.lowest_server_version.version)
        connection.close()

    def test_context_manager(self):
        with connect('localhost:4200') as conn:
            pass
        self.assertEqual(conn._closed, True)

    def test_with_timezone(self):
        """
        Verify the cursor objects will return timezone-aware `datetime` objects when requested to.
        When switching the time zone at runtime on the connection object, only new cursor objects
        will inherit the new time zone.
        """

        tz_mst = datetime.timezone(datetime.timedelta(hours=7), name="MST")
        connection = connect('localhost:4200', time_zone=tz_mst)
        cursor = connection.cursor()
        self.assertEqual(cursor.time_zone.tzname(None), "MST")
        self.assertEqual(cursor.time_zone.utcoffset(None), datetime.timedelta(seconds=25200))

        connection.time_zone = datetime.timezone.utc
        cursor = connection.cursor()
        self.assertEqual(cursor.time_zone.tzname(None), "UTC")
        self.assertEqual(cursor.time_zone.utcoffset(None), datetime.timedelta(0))

    def test_timeout_float(self):
        """
        Verify setting the timeout value as a scalar (float) works.
        """
        with connect('localhost:4200', timeout=2.42) as conn:
            self.assertEqual(conn.client._pool_kw["timeout"], 2.42)

    def test_timeout_string(self):
        """
        Verify setting the timeout value as a scalar (string) works.
        """
        with connect('localhost:4200', timeout="2.42") as conn:
            self.assertEqual(conn.client._pool_kw["timeout"], 2.42)

    def test_timeout_object(self):
        """
        Verify setting the timeout value as a Timeout object works.
        """
        timeout = Timeout(connect=2.42, read=0.01)
        with connect('localhost:4200', timeout=timeout) as conn:
            self.assertEqual(conn.client._pool_kw["timeout"], timeout)



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from __future__ import absolute_import
from unittest import TestCase, skipIf

import sqlalchemy as sa
from sqlalchemy.orm import Session
from sqlalchemy.sql.operators import eq

from crate.client.sqlalchemy import SA_VERSION, SA_1_4
from crate.testing.settings import crate_host

try:
    from sqlalchemy.orm import declarative_base
except ImportError:
    from sqlalchemy.ext.declarative import declarative_base

from crate.client.sqlalchemy.types import ObjectType, ObjectArray


class SqlAlchemyQueryCompilationCaching(TestCase):

    def setUp(self):
        self.engine = sa.create_engine(f"crate://{crate_host}")
        self.metadata = sa.MetaData(schema="testdrive")
        self.session = Session(bind=self.engine)
        self.Character = self.setup_entity()

    def setup_entity(self):
        """
        Define ORM entity.
        """
        Base = declarative_base(metadata=self.metadata)

        class Character(Base):
            __tablename__ = 'characters'
            name = sa.Column(sa.String, primary_key=True)
            age = sa.Column(sa.Integer)
            data = sa.Column(ObjectType)
            data_list = sa.Column(ObjectArray)

        return Character

    def setup_data(self):
        """
        Insert two records into the `characters` table.
        """
        self.metadata.drop_all(self.engine)
        self.metadata.create_all(self.engine)

        Character = self.Character
        char1 = Character(name='Trillian', data={'x': 1}, data_list=[{'foo': 1, 'bar': 10}])
        char2 = Character(name='Slartibartfast', data={'y': 2}, data_list=[{'bar': 2}])
        self.session.add(char1)
        self.session.add(char2)
        self.session.commit()
        self.session.execute(sa.text("REFRESH TABLE testdrive.characters;"))

    @skipIf(SA_VERSION < SA_1_4, "On SA13, the 'ResultProxy' object has no attribute 'scalar_one'")
    def test_object_multiple_select_legacy(self):
        """
        The SQLAlchemy implementation of CrateDB's `OBJECT` type offers indexed
        access to the instance's content in form of a dictionary. Thus, it must
        not use `cache_ok = True` on its implementation, i.e. this part of the
        compiled SQL clause must not be cached.

        This test verifies that two subsequent `SELECT` statements are translated
        well, and don't trip on incorrect SQL compiled statement caching.

        This variant uses direct value matching on the `OBJECT`s attribute.
        """
        self.setup_data()
        Character = self.Character

        selectable = sa.select(Character).where(Character.data['x'] == 1)
        result = self.session.execute(selectable).scalar_one().data
        self.assertEqual({"x": 1}, result)

        selectable = sa.select(Character).where(Character.data['y'] == 2)
        result = self.session.execute(selectable).scalar_one().data
        self.assertEqual({"y": 2}, result)

    @skipIf(SA_VERSION < SA_1_4, "On SA13, the 'ResultProxy' object has no attribute 'scalar_one'")
    def test_object_multiple_select_modern(self):
        """
        The SQLAlchemy implementation of CrateDB's `OBJECT` type offers indexed
        access to the instance's content in form of a dictionary. Thus, it must
        not use `cache_ok = True` on its implementation, i.e. this part of the
        compiled SQL clause must not be cached.

        This test verifies that two subsequent `SELECT` statements are translated
        well, and don't trip on incorrect SQL compiled statement caching.

        This variant uses comparator method matching on the `OBJECT`s attribute.
        """
        self.setup_data()
        Character = self.Character

        selectable = sa.select(Character).where(Character.data['x'].as_integer() == 1)
        result = self.session.execute(selectable).scalar_one().data
        self.assertEqual({"x": 1}, result)

        selectable = sa.select(Character).where(Character.data['y'].as_integer() == 2)
        result = self.session.execute(selectable).scalar_one().data
        self.assertEqual({"y": 2}, result)

    @skipIf(SA_VERSION < SA_1_4, "On SA13, the 'ResultProxy' object has no attribute 'scalar_one'")
    def test_objectarray_multiple_select(self):
        """
        The SQLAlchemy implementation of CrateDB's `ARRAY` type in form of the
        `ObjectArray`, does *not* offer indexed access to the instance's content.
        Thus, using `cache_ok = True` on that type should be sane, and not mess
        up SQLAlchemy's SQL compiled statement caching.
        """
        self.setup_data()
        Character = self.Character

        selectable = sa.select(Character).where(Character.data_list['foo'].any(1, operator=eq))
        result = self.session.execute(selectable).scalar_one().data
        self.assertEqual({"x": 1}, result)

        selectable = sa.select(Character).where(Character.data_list['bar'].any(2, operator=eq))
        result = self.session.execute(selectable).scalar_one().data
        self.assertEqual({"y": 2}, result)



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
import warnings

import sqlalchemy.types as sqltypes
from sqlalchemy.sql import operators, expression
from sqlalchemy.sql import default_comparator
from sqlalchemy.ext.mutable import Mutable

import geojson


class MutableList(Mutable, list):

    @classmethod
    def coerce(cls, key, value):
        """ Convert plain list to MutableList """
        if not isinstance(value, MutableList):
            if isinstance(value, list):
                return MutableList(value)
            elif value is None:
                return value
            else:
                return MutableList([value])
        else:
            return value

    def __init__(self, initval=None):
        list.__init__(self, initval or [])

    def __setitem__(self, key, value):
        list.__setitem__(self, key, value)
        self.changed()

    def __eq__(self, other):
        return list.__eq__(self, other)

    def append(self, item):
        list.append(self, item)
        self.changed()

    def insert(self, idx, item):
        list.insert(self, idx, item)
        self.changed()

    def extend(self, iterable):
        list.extend(self, iterable)
        self.changed()

    def pop(self, index=-1):
        list.pop(self, index)
        self.changed()

    def remove(self, item):
        list.remove(self, item)
        self.changed()


class MutableDict(Mutable, dict):

    @classmethod
    def coerce(cls, key, value):
        "Convert plain dictionaries to MutableDict."

        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def __init__(self, initval=None, to_update=None, root_change_key=None):
        initval = initval or {}
        self._changed_keys = set()
        self._deleted_keys = set()
        self._overwrite_key = root_change_key
        self.to_update = self if to_update is None else to_update
        for k in initval:
            initval[k] = self._convert_dict(initval[k],
                                            overwrite_key=k if self._overwrite_key is None else self._overwrite_key
                                            )
        dict.__init__(self, initval)

    def __setitem__(self, key, value):
        value = self._convert_dict(value, key if self._overwrite_key is None else self._overwrite_key)
        dict.__setitem__(self, key, value)
        self.to_update.on_key_changed(
            key if self._overwrite_key is None else self._overwrite_key
        )

    def __delitem__(self, key):
        dict.__delitem__(self, key)
        # add the key to the deleted keys if this is the root object
        # otherwise update on root object
        if self._overwrite_key is None:
            self._deleted_keys.add(key)
            self.changed()
        else:
            self.to_update.on_key_changed(self._overwrite_key)

    def on_key_changed(self, key):
        self._deleted_keys.discard(key)
        self._changed_keys.add(key)
        self.changed()

    def _convert_dict(self, value, overwrite_key):
        if isinstance(value, dict) and not isinstance(value, MutableDict):
            return MutableDict(value, self.to_update, overwrite_key)
        return value

    def __eq__(self, other):
        return dict.__eq__(self, other)


class ObjectTypeImpl(sqltypes.UserDefinedType, sqltypes.JSON):

    __visit_name__ = "OBJECT"

    cache_ok = False
    none_as_null = False


# Designated name to refer to. `Object` is too ambiguous.
ObjectType = MutableDict.as_mutable(ObjectTypeImpl)

# Backward-compatibility aliases.
_deprecated_Craty = ObjectType
_deprecated_Object = ObjectType

# https://www.lesinskis.com/deprecating-module-scope-variables.html
deprecated_names = ["Craty", "Object"]


def __getattr__(name):
    if name in deprecated_names:
        warnings.warn(f"{name} is deprecated and will be removed in future releases. "
                      f"Please use ObjectType instead.", DeprecationWarning)
        return globals()[f"_deprecated_{name}"]
    raise AttributeError(f"module {__name__} has no attribute {name}")


class Any(expression.ColumnElement):
    """Represent the clause ``left operator ANY (right)``.  ``right`` must be
    an array expression.

    copied from postgresql dialect

    .. seealso::

        :class:`sqlalchemy.dialects.postgresql.ARRAY`

        :meth:`sqlalchemy.dialects.postgresql.ARRAY.Comparator.any`
            ARRAY-bound method

    """
    __visit_name__ = 'any'
    inherit_cache = True

    def __init__(self, left, right, operator=operators.eq):
        self.type = sqltypes.Boolean()
        self.left = expression.literal(left)
        self.right = right
        self.operator = operator


class _ObjectArray(sqltypes.UserDefinedType):
    cache_ok = True

    class Comparator(sqltypes.TypeEngine.Comparator):
        def __getitem__(self, key):
            return default_comparator._binary_operate(self.expr,
                                                      operators.getitem,
                                                      key)

        def any(self, other, operator=operators.eq):
            """Return ``other operator ANY (array)`` clause.

            Argument places are switched, because ANY requires array
            expression to be on the right hand-side.

            E.g.::

                from sqlalchemy.sql import operators

                conn.execute(
                    select([table.c.data]).where(
                            table.c.data.any(7, operator=operators.lt)
                        )
                )

            :param other: expression to be compared
            :param operator: an operator object from the
             :mod:`sqlalchemy.sql.operators`
             package, defaults to :func:`.operators.eq`.

            .. seealso::

                :class:`.postgresql.Any`

                :meth:`.postgresql.ARRAY.Comparator.all`

            """
            return Any(other, self.expr, operator=operator)

    type = MutableList
    comparator_factory = Comparator

    def get_col_spec(self, **kws):
        return "ARRAY(OBJECT)"


ObjectArray = MutableList.as_mutable(_ObjectArray)


class Geopoint(sqltypes.UserDefinedType):
    cache_ok = True

    class Comparator(sqltypes.TypeEngine.Comparator):

        def __getitem__(self, key):
            return default_comparator._binary_operate(self.expr,
                                                      operators.getitem,
                                                      key)

    def get_col_spec(self):
        return 'GEO_POINT'

    def bind_processor(self, dialect):
        def process(value):
            if isinstance(value, geojson.Point):
                return value.coordinates
            return value
        return process

    def result_processor(self, dialect, coltype):
        return tuple

    comparator_factory = Comparator


class Geoshape(sqltypes.UserDefinedType):
    cache_ok = True

    class Comparator(sqltypes.TypeEngine.Comparator):

        def __getitem__(self, key):
            return default_comparator._binary_operate(self.expr,
                                                      operators.getitem,
                                                      key)

    def get_col_spec(self):
        return 'GEO_SHAPE'

    def result_processor(self, dialect, coltype):
        return geojson.GeoJSON.to_instance

    comparator_factory = Comparator



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
"""
Machinery for converting CrateDB database types to native Python data types.

https://crate.io/docs/crate/reference/en/latest/interfaces/http.html#column-types
"""
import ipaddress
from copy import deepcopy
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Union

ConverterFunction = Callable[[Optional[Any]], Optional[Any]]
ColTypesDefinition = Union[int, List[Union[int, "ColTypesDefinition"]]]


def _to_ipaddress(value: Optional[str]) -> Optional[Union[ipaddress.IPv4Address, ipaddress.IPv6Address]]:
    """
    https://docs.python.org/3/library/ipaddress.html
    """
    if value is None:
        return None
    return ipaddress.ip_address(value)


def _to_datetime(value: Optional[float]) -> Optional[datetime]:
    """
    https://docs.python.org/3/library/datetime.html
    """
    if value is None:
        return None
    return datetime.utcfromtimestamp(value / 1e3)


def _to_default(value: Optional[Any]) -> Optional[Any]:
    return value


# Symbolic aliases for the numeric data type identifiers defined by the CrateDB HTTP interface.
# https://crate.io/docs/crate/reference/en/latest/interfaces/http.html#column-types
class DataType(Enum):
    NULL = 0
    NOT_SUPPORTED = 1
    CHAR = 2
    BOOLEAN = 3
    TEXT = 4
    IP = 5
    DOUBLE = 6
    REAL = 7
    SMALLINT = 8
    INTEGER = 9
    BIGINT = 10
    TIMESTAMP_WITH_TZ = 11
    OBJECT = 12
    GEOPOINT = 13
    GEOSHAPE = 14
    TIMESTAMP_WITHOUT_TZ = 15
    UNCHECKED_OBJECT = 16
    REGPROC = 19
    TIME = 20
    OIDVECTOR = 21
    NUMERIC = 22
    REGCLASS = 23
    DATE = 24
    BIT = 25
    JSON = 26
    CHARACTER = 27
    ARRAY = 100


ConverterMapping = Dict[DataType, ConverterFunction]


# Map data type identifier to converter function.
_DEFAULT_CONVERTERS: ConverterMapping = {
    DataType.IP: _to_ipaddress,
    DataType.TIMESTAMP_WITH_TZ: _to_datetime,
    DataType.TIMESTAMP_WITHOUT_TZ: _to_datetime,
}


class Converter:
    def __init__(
        self,
        mappings: Optional[ConverterMapping] = None,
        default: ConverterFunction = _to_default,
    ) -> None:
        self._mappings = mappings or {}
        self._default = default

    def get(self, type_: ColTypesDefinition) -> ConverterFunction:
        if isinstance(type_, int):
            return self._mappings.get(DataType(type_), self._default)
        type_, inner_type = type_
        if DataType(type_) is not DataType.ARRAY:
            raise ValueError(f"Data type {type_} is not implemented as collection type")

        inner_convert = self.get(inner_type)

        def convert(value: Any) -> Optional[List[Any]]:
            if value is None:
                return None
            return [inner_convert(x) for x in value]

        return convert

    def set(self, type_: DataType, converter: ConverterFunction):
        self._mappings[type_] = converter


class DefaultTypeConverter(Converter):
    def __init__(self, more_mappings: Optional[ConverterMapping] = None) -> None:
        mappings: ConverterMapping = {}
        mappings.update(deepcopy(_DEFAULT_CONVERTERS))
        if more_mappings:
            mappings.update(deepcopy(more_mappings))
        super().__init__(
            mappings=mappings, default=_to_default
        )
